//
//  FLEXO II - Programa para clculo de propriedades de sees delgadas
//  Copyright (C) 1999  Maurcio C. Antunes (mantunes@sc.usp.br)
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//

#if !defined(pointsizerect_hpp)
#define pointsizerect_hpp

#include <vector>
#include <algorithm>

template <class T> class template_point;
template <class T> class template_size;
template <class T> class template_rect;

template <class T> class template_point {
public:
  // Coordenadas
  T x,y;

  // Constructors
  template_point() {};
  template_point(T _x, T _y) {x=_x;y=_y;};
  template_point(const template_point<T>& point){x=point.x;y=point.y;};

  // Information functions/operators
  bool IsNull() const {return x==0 && y==0;};
  bool operator ==(const template_point<T>& other) const {return other.x == x && other.y == y;};
  bool operator !=(const template_point<T>& other) const {return !(*this==other);};
  T X() const {return x;};
  T Y() const {return y;};
//  T Magnitude() const {return sqrt(x*x+y*y);};

  // Functions/binary-operators that return points or sizes
  template_point<T> OffsetBy(T dx, T dy) const {return template_point<T>(x + dx, y + dy);};
  template_point<T> operator +(const template_size<T>& size) const {return template_point<T>(x + size.cx, y + size.cy);};
  template_size<T> operator -(const template_point<T>& point) const {return template_size<T>(x - point.x, y - point.y);};
  template_point<T> operator -(const template_size<T>& size) const {return template_point<T>(x - size.cx, y - size.cy);};
  template_point<T> operator -() const {return template_point<T>(-x,-y);};

  // Functions/assignement-operators that modify this point
  template_point<T>& Offset(T dx, T dy) {x += dx;y += dy; return *this;};
  template_point<T>& operator +=(const template_size<T>& size){x += size.cx;y += size.cy;return *this;};
  template_point<T>& operator -=(const template_size<T>& size){x -= size.cx;y -= size.cy;return *this;};
};

template <class T> class template_size {
public:
  // Dimenses
  T cx,cy;

  // Constructors
  template_size<T>() {};
  template_size<T>(T dx, T dy) {cx = dx; cy = dy;};
  template_size<T>(const template_size<T>& size) {cx = size.cx; cy = size.cy;};

  // Information functions/operators
  bool IsNull() const {return cx==0 && cy==0;};
  bool operator ==(const template_size<T>& other) const {return other.cx==cx && other.cy==cy;};
  bool operator !=(const template_size<T>& other) const {return !(*this==other);};
  T X() const {return cx;};
  T Y() const {return cy;};
//  T Magnitude() const {return sqrt(cx*cx+cy*cy);};

  // Functions/binary-operators that return sizes
  template_size<T> operator +(const template_size<T>& size) const {return template_size<T>(cx+size.cx, cy+size.cy);};
  template_size<T> operator -(const template_size<T>& size) const {return template_size<T>(cx-size.cx, cy-size.cy);};
  template_size<T> operator -() const {return template_size<T>(-cx, -cy);};

  // Functions/assignement-operators that modify this size
  template_size<T>& operator +=(const template_size<T>& size) {cx += size.cx; cy += size.cy; return *this;};
  template_size<T>& operator -=(const template_size<T>& size) {cx -= size.cx; cy -= size.cy; return *this;};
};

template <class T> class template_rect {
public:
  // Coordenadas
  template_point<T> pi,pf;

  // Constructors
  template_rect() {};
  template_rect(const template_rect<T>& rect) {Set(rect.pi,rect.pf);};
  template_rect(T x1, T y1, T x2, T y2) {Set(x1,y1,x2,y2);};
  template_rect(const template_point<T>& p1, const template_point<T>& p2) {Set(p1,p2);};
  template_rect(const template_point<T>& origin, const template_size<T>& extent) {SetWH(origin,extent);};

  // (Re)Initializers
  void SetNull() {pi=pf=template_point<T>(0,0);};
  void SetEmpty() {SetNull();};
  void Set(T x1, T y1, T x2, T y2) {Set(template_point<T>(x1,y1),template_point<T>(x2,y2));};
  void Set(const template_point<T>& p1, const template_point<T>& p2){pi=p1;pf=p2;};
  void SetWH (const template_point<T>& origin, const template_size<T>& extent) {pi=pf=origin;pf+=extent;};

  // Testing functions
  bool IsEmpty() const {return pi.x==pf.x || pi.y==pf.y;};
  bool IsNull() const {return pi.IsNull() && pf.IsNull();};
  bool operator ==(const template_rect<T>& other) const {return other.pi==other.pi && other.pf==other.pf;};
  bool operator !=(const template_rect<T>& other) const {return !(*this==other);};

  // Information/access functions (const and non-const)
  T minX() const {return std::min(pi.x,pf.x);};
  T minY() const {return std::min(pi.y,pf.y);};
  T maxX() const {return std::max(pi.x,pf.x);};
  T maxY() const {return std::max(pi.y,pf.y);};

  template_point<T> minXminY() const {return template_point<T>(minX(),minY());};
  template_point<T> maxXmaxY() const {return template_point<T>(maxX(),maxY());};
  template_point<T> minXmaxY() const {return template_point<T>(minX(),maxY());};
  template_point<T> maxXminY() const {return template_point<T>(maxX(),minY());};
  T Width() const {return pf.x-pi.x;};
  T Height() const {return pf.y-pi.y;};
  template_size<T> Size() const {return template_size<T>(Width(),Height());};
  T Area() const {return Width()*Height();};

  bool Contains(const template_point<T>& point) const {return point.x >= minX() && point.x <= maxX() && point.y >= minY() && point.y <= maxY();};
  bool Contains(const template_rect<T>& other) const {return other.minX() >= minX() && other.maxX() <= maxX() && other.minY() >= minY() && other.maxY() <= maxY();};
  bool Touches(const template_rect<T>& other) const {return other.minX() <= maxX() && other.maxX() >= minX() && other.minY() <= maxY() && other.maxY() >= minY();};
  template_rect<T> OffsetBy(T dx, T dy) const {return (*this) + template_size<T>(dx,dy);};
  template_rect<T> operator +(const template_size<T>& size) const {return template_rect<T>(pi+size,pf+size);};
  template_rect<T> operator -(const template_size<T>& size) const {return template_rect<T>(pi-size,pf-size);};
  template_rect<T> MovedTo(T x, T y) {return template_rect<T>(x,y,x+Width(),y+Height());};
  template_rect<T> InflatedBy(T dx, T dy) const {return InflatedBy(template_size<T>(dx,dy));};
  template_rect<T> InflatedBy(const template_size<T>& size) const {return template_rect<T>(pi-size,pf+size);};
  template_rect<T> Normalized() const {return template_rect<T>(minXminY(),maxXmaxY());};
  template_rect<T> operator &(const template_rect<T>& other) const
    {
    if (!Touches(other)) return template_rect<T>(0,0,0,0);
    return template_rect<T> (
              std::max(minX(),other.minX()),
              std::max(minY(),other.minY()),
              std::min(maxX(),other.maxX()),
              std::min(maxY(),other.maxY())
              );
    };
  template_rect<T> operator |(const template_rect<T>& other) const
    {
    return template_rect<T>(
              std::min(minX(),other.minX()),
              std::min(minY(),other.minY()),
              std::max(maxX(),other.maxX()),
              std::max(maxY(),other.maxY())
              );
    };

//  void Subtract(const template_rect<T>& other, std::vector<template_rect<T> >& result) const{};

  // Manipulation functions/operators
  template_rect<T>& Normalize() {Set(minXminY(),maxXmaxY());return *this;};
  template_rect<T>& Offset(T dx, T dy) {(*this)+=(template_size<T>(dx,dy));return *this;};
  template_rect<T>& operator +=(const template_size<T>& delta) {pi+=delta; pf+=delta; return *this;};
  template_rect<T>& operator -=(const template_size<T>& delta) {pi-=delta; pf-=delta; return *this;};
  template_rect<T>& MoveTo(T x, T y) {Set(x,y,x+Width(),y+Height()); return *this;};
  template_rect<T>& Inflate(T dx, T dy) {Inflate(template_size<T>(dx,dy)); return *this;};
  template_rect<T>& Inflate(const template_size<T>& delta) {Set(pi-delta,pf+delta);return *this;};
  template_rect<T>& operator &=(const template_rect<T>& other)
    {
    if (!Touches(other)) Set(0,0,0,0);
    else Set(
              std::max(minX(),other.minX()),
              std::max(minY(),other.minY()),
              std::min(maxX(),other.maxX()),
              std::min(maxY(),other.maxY())
              );
    return *this;
    };
  template_rect<T>& operator |=(const template_rect<T>& other)
    {
    Set(
              std::min(minX(),other.minX()),
              std::min(minY(),other.minY()),
              std::max(maxX(),other.maxX()),
              std::max(maxY(),other.maxY())
              );
    return *this;
    };
};




#endif //pointsizerect_hpp
